package edu.rice.comp322;

import edu.rice.hj.api.HjRunnable;
import edu.rice.hj.api.HjSuspendable;
import edu.rice.hj.api.SuspendableException;
import edu.rice.hj.runtime.config.HjConfiguration;
import edu.rice.hj.runtime.config.HjSystemProperty;
import junit.framework.TestCase;

import static edu.rice.comp322.OneDimAveraging.setOutput;
import static edu.rice.hj.Module0.finish;
import static edu.rice.hj.Module0.launchHabaneroApp;

/**
 * Simple Unit test.
 */
public class OneDimAveragingPerformanceTest extends TestCase {

    private final int size = 1_600_000;
    private final int iterations = 500;
    private final int runs = 5;

    public void setUp() throws Exception {
        OneDimAveraging.initialOutput = null;
        final OneDimAveraging seqSolution = new OneDimAveraging(size);

        launchHabaneroApp(() -> {
            seqSolution.runSequential(iterations);
            setOutput(seqSolution);
        });
    }

    public void testAveragingUsingParallel() {
        System.out.println("OneDimAveragingTest.testAveragingUsingParallel() starts...");

        launchHabaneroApp(() -> {
            final OneDimAveraging[] solution = {null};
            final long meanTime = timeIt(runs,
                () -> {
                    solution[0] = new OneDimAveraging(size);
                },
                () -> {
                    solution[0].runSequential(iterations);
                }, () -> {
                    solution[0] = null;
                });
            System.out.println(" Average sequential time = " + meanTime + " ms.");
        });

        final int numTasks = HjConfiguration.readIntProperty(HjSystemProperty.numWorkers);
        System.out.println("OneDimAveragingTest: numTasks = " + numTasks);

        launchHabaneroApp(() -> {
            final OneDimAveraging[] solution = {null};
            final long meanTime = timeIt(runs,
                () -> {
                    solution[0] = new OneDimAveraging(size);
                },
                () -> {
                    solution[0].runChunkedPhaser(iterations, numTasks);
                }, () -> {
                    assertTrue("ERROR: Sequential and Parallel output are not same!", solution[0].validateOutput());
                    solution[0] = null;
                });
            System.out.println(" Passed! Average Chunked-Phaser time = " + meanTime + " ms.");
        });

        launchHabaneroApp(() -> {
            final OneDimAveraging[] solution = {null};
            final long meanTime = timeIt(runs,
                () -> {
                    solution[0] = new OneDimAveraging(size);
                },
                () -> {
                    solution[0].runChunkedBarrier(iterations, numTasks);
                }, () -> {
                    assertTrue("ERROR: Sequential and Parallel output are not same!", solution[0].validateOutput());
                    solution[0] = null;
                });
            System.out.println(" Passed! Average Chunked-Barrier time = " + meanTime + " ms.");
        });

        System.out.println("OneDimAveragingTest.testAveragingUsingParallel() ends.");
    }

    private static long timeIt(
        final int runs,
        final HjRunnable prolog,
        final HjSuspendable actualBody,
        final HjRunnable epilog)
        throws SuspendableException {

        long aggregateTime = 0;

        for (int r = 0; r <= runs; r++) {
            prolog.run();
            final long s = System.currentTimeMillis();
            finish(() -> {
                actualBody.run();
            });
            final long e = System.currentTimeMillis();
            if (r > 0) {
                final long iterTime = e - s;
                System.out.println("  Iteration-" + r + " time = " + iterTime + " ms.");
                aggregateTime += iterTime;
            }
            epilog.run();
        }

        return (aggregateTime / runs);
    }
}
